package modeler.literals.helpers;

import modeler.CNFFormula;
import modeler.CNFFormula.Clause;
import java.util.*;
import modeler.literals.IntegerTuple;
import modeler.literals.TupleMap;
import modeler.literals.helpers.RangeSettings.RangeSetting;
import modeler.literals.helpers.RangeSettings.Range.RangeIterator;

public class TupleLiteralHelper {

	public static void createMutuallyExclusiveClauses(TupleMap map, CNFFormula formula, RangeSettings settings) {
		Vector<RangeSetting> fixed = new Vector<RangeSetting>();
		for (Iterator<RangeSetting> i = settings.getFixedSettings(); i.hasNext(); ) {
			fixed.add(i.next());
		}

		Vector<RangeSetting> variable = new Vector<RangeSetting>();
		for (Iterator<RangeSetting> i = settings.getVariableSettings(); i.hasNext(); ) {
			variable.add(i.next());
		}

		try {
			FixedPointsGenerator fpg = new MutuallyExclusiveClausesGenerator(fixed, variable, map, formula);
			fpg.generatePoints();
		} catch(GenerationException e) {
			assert false : "Should not happen.";
		}
	}

	public static void createORClauses(TupleMap map, CNFFormula formula, RangeSettings settings) {
		Vector<RangeSetting> fixed = new Vector<RangeSetting>();
		for (Iterator<RangeSetting> i = settings.getFixedSettings(); i.hasNext(); ) {
			fixed.add(i.next());
		}

		Vector<RangeSetting> variable = new Vector<RangeSetting>();
		for (Iterator<RangeSetting> i = settings.getVariableSettings(); i.hasNext(); ) {
			variable.add(i.next());
		}

		try {
			FixedPointsGenerator fpg = new ORClausesGenerator(fixed, variable, map, formula);
			fpg.generatePoints();
		} catch(GenerationException e) {
			assert false : "Should not happen.";
		}
	}

	public static class PointsGenerator {

		public PointsGenerator(Vector<RangeSetting> settings) {
			this.settings = settings;
		}

		public void generatePoints() {
			/* Needed variables */
			int i, c = this.settings.size();
			RangeIterator it;

			/* Initialize the array of values and iterators. */
			RangeIterator[] iterators = new RangeIterator[this.settings.size()];
			int[] values = new int[this.settings.size()];
			for (i = 0; i < c; ++i) {
				iterators[i] = this.settings.get(i).getRange().rangeIterator();
				values[i] = iterators[i].next();
			}

			/* Generate points */
			while (true) {
				/* Register a point */
				this.onPointGenerated(values);

				/* Move to the next point in the set. */
				for (i = c - 1; i >= 0; --i) {
					it = iterators[i];

					if (it.hasNext()) {
						values[i] = it.next();
						break;
					} else {
						it.reset();
						values[i] = it.next();
					}
				}

				/* No other possibility */
				if (i == -1) {
					return;
				}
			}
		}

		protected void onPointGenerated(int[] point) {
			assert point.length == this.settings.size() : "Lengths do not match.";
		}

		protected final Vector<RangeSetting> settings;

	}

	public static class GenerationException extends Exception {

		private static final long serialVersionUID = 1L;

		public GenerationException(String message) {
			super(message);
		}

	}

	public static class VariablePointsGenerator extends PointsGenerator {

		public VariablePointsGenerator(Vector<RangeSetting> variableSettings) throws GenerationException {
			super(variableSettings);

			for (RangeSetting s : variableSettings) {
				if (s.isFixed()) {
					throw new GenerationException("Invalid variable settings.");
				}
			}
		}

		protected void onPointGenerated(int[] point) {
			super.onPointGenerated(point);
			int[] cpy = new int[point.length];
			for (int i = 0, c = point.length; i != c; ++i) {
				cpy[i] = point[i];
			}

			this.variablePoints.add(cpy);
		}

		protected Vector<int[]> getVariablePoints() {
			return this.variablePoints;
		}

		protected final Vector<int[]> variablePoints = new Vector<int[]>();

	}

	public static class FixedPointsGenerator extends PointsGenerator {

		public FixedPointsGenerator(Vector<RangeSetting> fixedSettings, Vector<RangeSetting> variableSettings, TupleMap map, CNFFormula formula)
				throws GenerationException {
			super(fixedSettings);

			for (RangeSetting s : fixedSettings) {
				if (!s.fixed) {
					throw new GenerationException("Invalid fixed settings.");
				}
			}

			VariablePointsGenerator vpg = new VariablePointsGenerator(variableSettings);
			vpg.generatePoints();

			this.variablePoints = vpg.getVariablePoints();
			this.variableSettings = variableSettings;
			this.map = map;
			this.formula = formula;
		}

		protected void fillPoint(int[] resPoint, int[] point, Vector<RangeSetting> settings) {
			for (int i = 0, c = settings.size(); i != c; ++i) {
				resPoint[settings.get(i).getDimension()] = point[i];
			}
		}

		protected final Vector<int[]> variablePoints;

		protected final Vector<RangeSetting> variableSettings;

		protected final TupleMap map;

		protected final CNFFormula formula;

	}

	public static class MutuallyExclusiveClausesGenerator extends FixedPointsGenerator {

		public MutuallyExclusiveClausesGenerator(Vector<RangeSetting> fixedSettings, Vector<RangeSetting> variableSettings,
				TupleMap map, CNFFormula formula) throws GenerationException {

			super(fixedSettings, variableSettings, map, formula);
		}

		protected void onPointGenerated(int[] point) {
			super.onPointGenerated(point);

			int[] myPoint1 = new int[this.settings.size() + this.variableSettings.size()];
			int[] myPoint2 = new int[this.settings.size() + this.variableSettings.size()];
			for (int i = 0, c = this.variablePoints.size(); i != c; ++i) {
				for (int j = i + 1; j != c; ++j) {
					// Assemble the points.
					this.fillPoint(myPoint1, this.variablePoints.get(i), this.variableSettings);
					this.fillPoint(myPoint2, this.variablePoints.get(j), this.variableSettings);
					this.fillPoint(myPoint1, point, this.settings);
					this.fillPoint(myPoint2, point, this.settings);

					// Make the clause.
					Clause clause = this.formula.new Clause();
					clause.addLiteral(map.getTupleLiteral(false, new IntegerTuple(myPoint1)));
					clause.addLiteral(map.getTupleLiteral(false, new IntegerTuple(myPoint2)));
					this.formula.addClause(clause);
				}
			}
		}

	}

	public static class ORClausesGenerator extends FixedPointsGenerator {

		public ORClausesGenerator(Vector<RangeSetting> fixedSettings, Vector<RangeSetting> variableSettings,
				TupleMap map, CNFFormula formula) throws GenerationException {

			super(fixedSettings, variableSettings, map, formula);
		}

		protected void onPointGenerated(int[] point) {
			super.onPointGenerated(point);

			int[] myPoint = new int[this.settings.size() + this.variableSettings.size()];
			Clause clause = this.formula.new Clause();

			// For all variables create the clause.
			for (int i = 0, c = this.variablePoints.size(); i != c; ++i) {
				// Assemble the point.
				this.fillPoint(myPoint, this.variablePoints.get(i), this.variableSettings);
				this.fillPoint(myPoint, point, this.settings);

				// Add to the clause.
				clause.addLiteral(map.getTupleLiteral(true, new IntegerTuple(myPoint)));
			}

			this.formula.addClause(clause);
		}

	}

}
